from __future__ import annotations

import asyncio
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, assert_never, overload

from opentelemetry import trace as otel_trace

from any_agent.config import (
    AgentConfig,
    AgentFramework,
    Tool,
)
from any_agent.logging import logger
from any_agent.tools.wrappers import _wrap_tools
from any_agent.tracing.agent_trace import AgentTrace
from any_agent.tracing.exporter import SCOPE_NAME
from any_agent.tracing.instrumentation import (
    _get_instrumentor_by_framework,
)
from any_agent.utils import run_async_in_sync

if TYPE_CHECKING:
    from collections.abc import Sequence

    from opentelemetry.trace import Tracer
    from pydantic import BaseModel

    from any_agent.serving import A2AServingConfig, MCPServingConfig, ServerHandle
    from any_agent.tools.mcp.mcp_server import _MCPServerBase


class AgentRunError(Exception):
    """Error that wraps underlying framework specific errors and carries spans."""

    _trace: AgentTrace
    _original_exception: Exception

    def __init__(self, trace: AgentTrace, original_exception: Exception):
        self._trace = trace
        self._original_exception = original_exception
        # Set the exception message to be the original exception's message
        super().__init__(str(original_exception))

    @property
    def trace(self) -> AgentTrace:
        return self._trace

    @property
    def original_exception(self) -> Exception:
        return self._original_exception

    def __str__(self) -> str:
        """Return the string representation of the original exception."""
        return str(self._original_exception)

    def __repr__(self) -> str:
        """Return the detailed representation of the AgentRunError."""
        return f"AgentRunError({self._original_exception!r})"


class AnyAgent(ABC):
    """Base abstract class for all agent implementations.

    This provides a unified interface for different agent frameworks.
    """

    def __init__(self, config: AgentConfig):
        self.config = config

        self._mcp_servers: list[_MCPServerBase[Any]] = []
        self._tools: list[Any] = []

        self._instrumentor = _get_instrumentor_by_framework(self.framework)
        self._tracer: Tracer = otel_trace.get_tracer(SCOPE_NAME)

        self._lock = asyncio.Lock()
        self._running_traces: dict[int, AgentTrace] = {}

    @staticmethod
    def _get_agent_type_by_framework(
        framework_raw: AgentFramework | str,
    ) -> type[AnyAgent]:
        framework = AgentFramework.from_string(framework_raw)

        if framework is AgentFramework.SMOLAGENTS:
            from any_agent.frameworks.smolagents import SmolagentsAgent

            return SmolagentsAgent

        if framework is AgentFramework.LANGCHAIN:
            from any_agent.frameworks.langchain import LangchainAgent

            return LangchainAgent

        if framework is AgentFramework.OPENAI:
            from any_agent.frameworks.openai import OpenAIAgent

            return OpenAIAgent

        if framework is AgentFramework.LLAMA_INDEX:
            from any_agent.frameworks.llama_index import LlamaIndexAgent

            return LlamaIndexAgent

        if framework is AgentFramework.GOOGLE:
            from any_agent.frameworks.google import GoogleAgent

            return GoogleAgent

        if framework is AgentFramework.AGNO:
            from any_agent.frameworks.agno import AgnoAgent

            return AgnoAgent

        if framework is AgentFramework.TINYAGENT:
            from any_agent.frameworks.tinyagent import TinyAgent

            return TinyAgent

        assert_never(framework)

    @classmethod
    def create(
        cls,
        agent_framework: AgentFramework | str,
        agent_config: AgentConfig,
    ) -> AnyAgent:
        """Create an agent using the given framework and config."""
        return run_async_in_sync(
            cls.create_async(
                agent_framework=agent_framework,
                agent_config=agent_config,
            )
        )

    @classmethod
    async def create_async(
        cls,
        agent_framework: AgentFramework | str,
        agent_config: AgentConfig,
    ) -> AnyAgent:
        """Create an agent using the given framework and config."""
        agent_cls = cls._get_agent_type_by_framework(agent_framework)
        agent = agent_cls(agent_config)
        await agent._load_agent()
        return agent

    async def _load_tools(
        self, tools: Sequence[Tool]
    ) -> tuple[list[Any], list[_MCPServerBase[Any]]]:
        tools, mcp_servers = await _wrap_tools(tools, self.framework)
        # Add to agent so that it doesn't get garbage collected
        self._mcp_servers.extend(mcp_servers)
        for mcp_server in mcp_servers:
            tools.extend(mcp_server.tools)
        return tools, mcp_servers

    def run(self, prompt: str, **kwargs: Any) -> AgentTrace:
        """Run the agent with the given prompt."""
        return run_async_in_sync(self.run_async(prompt, **kwargs))

    async def run_async(
        self, prompt: str, instrument: bool = True, **kwargs: Any
    ) -> AgentTrace:
        """Run the agent asynchronously with the given prompt.

        Args:
            prompt: The user prompt to be passed to the agent.
            instrument: Whether to instrument the underlying framework
                to generate LLM Calls and Tool Execution Spans.

                If `False` the returned `AgentTrace` will only
                contain a single `invoke_agent` span.

            kwargs: Will be passed to the underlying runner used
                by the framework.

        Returns:
            The `AgentTrace` containing information about the
                steps taken by the agent.

        """
        trace = AgentTrace()
        trace_id: int
        instrumentation_enabled = instrument and self._instrumentor is not None

        # This design is so that we only catch exceptions thrown by _run_async. All other exceptions will not be caught.
        try:
            with self._tracer.start_as_current_span(
                f"invoke_agent [{self.config.name}]"
            ) as invoke_span:
                if instrumentation_enabled:
                    trace_id = invoke_span.get_span_context().trace_id
                    async with self._lock:
                        # We check the locked `_running_traces` inside `instrument`.
                        # If there is more than 1 entry in `running_traces`, it means that the agent has
                        # already being instrumented so we won't instrument it again.
                        self._running_traces[trace_id] = AgentTrace()
                        self._instrumentor.instrument(
                            agent=self,  # type: ignore[arg-type]
                        )

                invoke_span.set_attributes(
                    {
                        "gen_ai.operation.name": "invoke_agent",
                        "gen_ai.agent.name": self.config.name,
                        "gen_ai.agent.description": self.config.description
                        or "No description.",
                        "gen_ai.request.model": self.config.model_id,
                    }
                )
                final_output = await self._run_async(prompt, **kwargs)
        except Exception as e:
            # Clean up instrumentation if it was enabled
            if instrumentation_enabled:
                async with self._lock:
                    self._instrumentor.uninstrument(self)  # type: ignore[arg-type]
                    # Get the instrumented trace if available, otherwise use the original trace
                    instrumented_trace = self._running_traces.pop(trace_id)
                    if instrumented_trace is not None:
                        trace = instrumented_trace
            trace.add_span(invoke_span)
            raise AgentRunError(trace, e) from e

        if instrumentation_enabled:
            async with self._lock:
                self._instrumentor.uninstrument(self)  # type: ignore[arg-type]
                trace = self._running_traces.pop(trace_id)

        trace.add_span(invoke_span)
        trace.final_output = final_output
        return trace

    def _serve_a2a(self, serving_config: A2AServingConfig | None) -> None:
        from any_agent.serving import A2AServingConfig, _get_a2a_app, serve_a2a

        if serving_config is None:
            serving_config = A2AServingConfig()

        app = _get_a2a_app(self, serving_config=serving_config)

        serve_a2a(
            app,
            host=serving_config.host,
            port=serving_config.port,
            endpoint=serving_config.endpoint,
            log_level=serving_config.log_level,
        )

    def _serve_mcp(self, serving_config: MCPServingConfig) -> None:
        from any_agent.serving import (
            serve_mcp,
        )

        serve_mcp(
            self,
            host=serving_config.host,
            port=serving_config.port,
            endpoint=serving_config.endpoint,
            log_level=serving_config.log_level,
        )

    @overload
    def serve(self, serving_config: MCPServingConfig) -> None: ...

    @overload
    def serve(self, serving_config: A2AServingConfig | None = None) -> None: ...

    def serve(
        self, serving_config: MCPServingConfig | A2AServingConfig | None = None
    ) -> None:
        """Serve this agent using the protocol defined in the serving_config.

        Args:
            serving_config: Configuration for serving the agent. If None, uses default A2AServingConfig.
                          Must be an instance of A2AServingConfig or MCPServingConfig.

        Raises:
            ImportError: If the `a2a` dependencies are not installed and an `A2AServingConfig` is used.

        Example:
            ```
            agent = AnyAgent.create("tinyagent", AgentConfig(...))
            config = A2AServingConfig(port=8080, endpoint="/my-agent")
            agent.serve(config)
            ```

        """
        from any_agent.serving import MCPServingConfig

        if isinstance(serving_config, MCPServingConfig):
            self._serve_mcp(serving_config)
        else:
            self._serve_a2a(serving_config)

    async def _serve_a2a_async(
        self, serving_config: A2AServingConfig | None
    ) -> ServerHandle:
        from any_agent.serving import (
            A2AServingConfig,
            _get_a2a_app_async,
            serve_a2a_async,
        )

        if serving_config is None:
            serving_config = A2AServingConfig()

        app = await _get_a2a_app_async(self, serving_config=serving_config)

        return await serve_a2a_async(
            app,
            host=serving_config.host,
            port=serving_config.port,
            endpoint=serving_config.endpoint,
            log_level=serving_config.log_level,
        )

    async def _serve_mcp_async(self, serving_config: MCPServingConfig) -> ServerHandle:
        from any_agent.serving import serve_mcp_async

        return await serve_mcp_async(
            self,
            host=serving_config.host,
            port=serving_config.port,
            endpoint=serving_config.endpoint,
            log_level=serving_config.log_level,
        )

    @overload
    async def serve_async(self, serving_config: MCPServingConfig) -> ServerHandle: ...

    @overload
    async def serve_async(
        self, serving_config: A2AServingConfig | None = None
    ) -> ServerHandle: ...

    async def serve_async(
        self, serving_config: MCPServingConfig | A2AServingConfig | None = None
    ) -> ServerHandle:
        """Serve this agent asynchronously using the protocol defined in the serving_config.

        Args:
            serving_config: Configuration for serving the agent. If None, uses default A2AServingConfig.
                          Must be an instance of A2AServingConfig or MCPServingConfig.

        Returns:
            A ServerHandle instance that provides methods for managing the server lifecycle.

        Raises:
            ImportError: If the `a2a` dependencies are not installed and an `A2AServingConfig` is used.

        Example:
            ```
            agent = await AnyAgent.create_async("tinyagent", AgentConfig(...))
            config = MCPServingConfig(port=8080)
            server_handle = await agent.serve_async(config)
            try:
                # Server is running
                await asyncio.sleep(10)
            finally:
                await server_handle.shutdown()
            ```

        """
        from any_agent.serving import MCPServingConfig

        if isinstance(serving_config, MCPServingConfig):
            return await self._serve_mcp_async(serving_config)
        return await self._serve_a2a_async(serving_config)

    def _recreate_with_config(self, new_config: AgentConfig) -> AnyAgent:
        """Create a new agent instance with the given config, preserving MCP servers and tools.

        This method creates a new agent with the modified configuration while transferring
        the MCP servers and tools from the current agent to avoid recreating them, but only
        if the tools configuration hasn't changed.

        Args:
            new_config: The new configuration to use for the recreated agent.

        Returns:
            A new agent instance with the modified config and transferred state (if tools unchanged)
            or a completely new agent (if tools changed).

        """
        return run_async_in_sync(self._recreate_with_config_async(new_config))

    async def _recreate_with_config_async(self, new_config: AgentConfig) -> AnyAgent:
        """Async version of _recreate_with_config.

        This method creates a new agent with the modified configuration while transferring
        the MCP servers and tools from the current agent to avoid recreating them, but only
        if the tools configuration hasn't changed.

        Args:
            new_config: The new configuration to use for the recreated agent.

        Returns:
            A new agent instance with the modified config and transferred state (if tools unchanged)
            or a completely new agent (if tools changed).

        """
        # Check if tools configuration has changed
        if self.config.tools != new_config.tools:
            # Tools have changed, so we need to recreate everything from scratch
            logger.info(
                "Tools have changed, so we need to recreate everything from scratch"
            )
            return await self.create_async(self.framework, new_config)

        # Tools haven't changed, so we can safely preserve MCP servers and tools
        # Create the new agent with the modified config
        # Don't use AnyAgent.create_async(), because it will recreate the MCP servers and tools
        new_agent = self.__class__(new_config)

        # Transfer MCP servers and tools from the original agent to avoid recreating them
        new_agent._mcp_servers = self._mcp_servers
        new_agent._tools = self._tools

        # Load the agent with the existing MCP servers and tools
        await new_agent._load_agent()

        return new_agent

    @abstractmethod
    async def _load_agent(self) -> None:
        """Load the agent instance."""

    @abstractmethod
    async def _run_async(self, prompt: str, **kwargs: Any) -> str | BaseModel:
        """To be implemented by each framework."""

    @property
    @abstractmethod
    def framework(self) -> AgentFramework:
        """The Agent Framework used."""

    @property
    def agent(self) -> Any:
        """The underlying agent implementation from the framework.

        This property is intentionally restricted to maintain framework abstraction
        and prevent direct dependency on specific agent implementations.

        If you need functionality that relies on accessing the underlying agent:
        1. Consider if the functionality can be added to the AnyAgent interface
        2. Submit a GitHub issue describing your use case
        3. Contribute a PR implementing the needed functionality

        Raises:
            NotImplementedError: Always raised when this property is accessed

        """
        msg = "Cannot access the 'agent' property of AnyAgent, if you need to use functionality that relies on the underlying agent framework, please file a Github Issue or we welcome a PR to add the functionality to the AnyAgent class"
        raise NotImplementedError(msg)



from unittest.mock import MagicMock

from google.adk.models.llm_request import LlmRequest
from google.adk.models.llm_response import LlmResponse

from any_agent.tracing.instrumentation.google import (
    _GoogleADKInstrumentor,
    _set_llm_input,
    _set_llm_output,
)


def test_set_llm_input_missing_fields() -> None:
    """It should not fail when missing fields."""
    span = MagicMock()
    _set_llm_input(LlmRequest(), span)

    span.set_attribute.assert_not_called()


def test_set_llm_output_missing_fields() -> None:
    """It should not fail when missing fields."""
    span = MagicMock()
    _set_llm_output(LlmResponse(), span)

    span.set_attributes.assert_not_called()


def test_uninstrument_before_instrument() -> None:
    _GoogleADKInstrumentor().uninstrument(MagicMock())


def test_instrument_uninstrument() -> None:
    """Regression test for https://github.com/mozilla-ai/any-agent/issues/467"""
    agent = MagicMock()
    agent._agent.before_model_callback = None
    agent._agent.after_model_callback = None
    agent._agent.before_tool_callback = None
    agent._agent.after_tool_callback = None
    instrumentor = _GoogleADKInstrumentor()

    instrumentor.instrument(agent)
    assert callable(agent._agent.before_model_callback)
    assert callable(agent._agent.after_model_callback)
    assert callable(agent._agent.before_tool_callback)
    assert callable(agent._agent.after_tool_callback)

    instrumentor.uninstrument(agent)
    assert agent._agent.before_model_callback is None
    assert agent._agent.after_model_callback is None
    assert agent._agent.before_tool_callback is None
    assert agent._agent.after_tool_callback is None



"""MCP adapter for Tiny framework."""

import os
from abc import ABC, abstractmethod
from collections.abc import Callable, Sequence
from contextlib import suppress
from datetime import timedelta
from typing import Any, Literal

from pydantic import Field, PrivateAttr

from any_agent.config import AgentFramework, MCPSse, MCPStdio, Tool
from any_agent.tools.mcp.mcp_connection import _MCPConnection
from any_agent.tools.mcp.mcp_server import _MCPServerBase

# Check for MCP dependencies
mcp_available = False
with suppress(ImportError):
    from mcp import ClientSession, StdioServerParameters
    from mcp.client.sse import sse_client
    from mcp.client.stdio import stdio_client
    from mcp.types import Tool as MCPTool  # noqa: TC002

    mcp_available = True


class TinyAgentMCPConnection(_MCPConnection["MCPTool"], ABC):
    """Base class for TinyAgent MCP connections."""

    _client: Any | None = PrivateAttr(default=None)

    @abstractmethod
    async def list_tools(self) -> list["MCPTool"]:
        """List tools from the MCP server."""
        if not self._client:
            msg = "MCP client is not set up. Please call `setup` from a concrete class."
            raise ValueError(msg)

        # Setup the client connection using exit stack to manage lifecycle
        stdio, write = await self._exit_stack.enter_async_context(self._client)

        # Create a client session
        client_session = ClientSession(
            stdio,
            write,
            timedelta(seconds=self.mcp_tool.client_session_timeout_seconds)
            if self.mcp_tool.client_session_timeout_seconds
            else None,
        )

        # Start the session
        session: ClientSession = await self._exit_stack.enter_async_context(
            client_session
        )
        if not session:
            msg = "Failed to create MCP session"
            raise ValueError(msg)

        await session.initialize()

        # Get the available tools from the MCP server using schema
        available_tools = await session.list_tools()

        # Filter tools if specific tools were requested
        filtered_tools = self._filter_tools(available_tools.tools)

        # Create callable tool functions
        tool_list = list[Any]()
        for tool_info in filtered_tools:
            tool_list.append(self._create_tool_from_info(tool_info, session))  # type: ignore[arg-type]

        return tool_list

    def _create_tool_from_info(
        self, tool: Tool, session: "ClientSession"
    ) -> Callable[..., Any]:
        """Create a tool function from tool information."""
        tool_name = tool.name if hasattr(tool, "name") else tool
        tool_description = tool.description if hasattr(tool, "description") else ""
        input_schema = tool.inputSchema if hasattr(tool, "inputSchema") else None
        if not session:
            msg = "Not connected to MCP server"
            raise ValueError(msg)

        async def tool_function(*args, **kwargs) -> Any:  # type: ignore[no-untyped-def]
            """Tool function that calls the MCP server."""
            # Combine args and kwargs
            combined_args = {}
            if args and len(args) > 0:
                combined_args = args[0]
            combined_args.update(kwargs)

            if not session:
                msg = "Not connected to MCP server"
                raise ValueError(msg)
            # Call the tool on the MCP server
            try:
                return await session.call_tool(tool_name, combined_args)  # type: ignore[arg-type]
            except Exception as e:
                return f"Error calling tool {tool_name}: {e!s}"

        # Set attributes for the tool function
        tool_function.__name__ = tool_name  # type: ignore[assignment]
        tool_function.__doc__ = tool_description
        # this isn't a defined attribute of a callable, but we pass it to tinyagent so that it can use it appropriately
        # when constructing the ToolExecutor.
        tool_function.__input_schema__ = input_schema  # type: ignore[attr-defined]

        return tool_function


class TinyAgentMCPStdioConnection(TinyAgentMCPConnection):
    mcp_tool: MCPStdio

    async def list_tools(self) -> list["MCPTool"]:
        """List tools from the MCP server."""
        server_params = StdioServerParameters(
            command=self.mcp_tool.command,
            args=list(self.mcp_tool.args),
            env={**os.environ},
        )

        self._client = stdio_client(server_params)

        return await super().list_tools()


class TinyAgentMCPSseConnection(TinyAgentMCPConnection):
    mcp_tool: MCPSse

    async def list_tools(self) -> list["MCPTool"]:
        """List tools from the MCP server."""
        kwargs = {}
        if self.mcp_tool.client_session_timeout_seconds:
            kwargs["sse_read_timeout"] = self.mcp_tool.client_session_timeout_seconds
        self._client = sse_client(
            url=self.mcp_tool.url,
            headers=dict(self.mcp_tool.headers or {}),
            **kwargs,  # type: ignore[arg-type]
        )

        return await super().list_tools()


class TinyAgentMCPServerBase(_MCPServerBase["MCPTool"], ABC):
    framework: Literal[AgentFramework.TINYAGENT] = AgentFramework.TINYAGENT
    tools: Sequence["MCPTool"] = Field(default_factory=list)
    libraries: str = "any-agent[mcp]"

    def _check_dependencies(self) -> None:
        """Check if the required dependencies for the MCP server are available."""
        self.mcp_available = mcp_available
        super()._check_dependencies()


class TinyAgentMCPServerStdio(TinyAgentMCPServerBase):
    mcp_tool: MCPStdio

    async def _setup_tools(
        self, mcp_connection: _MCPConnection["MCPTool"] | None = None
    ) -> None:
        mcp_connection = mcp_connection or TinyAgentMCPStdioConnection(
            mcp_tool=self.mcp_tool
        )
        await super()._setup_tools(mcp_connection)


class TinyAgentMCPServerSse(TinyAgentMCPServerBase):
    mcp_tool: MCPSse

    async def _setup_tools(
        self, mcp_connection: _MCPConnection["MCPTool"] | None = None
    ) -> None:
        mcp_connection = mcp_connection or TinyAgentMCPSseConnection(
            mcp_tool=self.mcp_tool
        )
        await super()._setup_tools(mcp_connection)


TinyAgentMCPServer = TinyAgentMCPServerStdio | TinyAgentMCPServerSse



from unittest.mock import MagicMock

from smolagents.models import ChatMessage

from any_agent.tracing.instrumentation.smolagents import (
    _set_llm_input,
    _set_llm_output,
    _SmolagentsInstrumentor,
)


def test_set_llm_input_missing_fields() -> None:
    """It should not fail when missing fields."""
    span = MagicMock()
    _set_llm_input([], span)

    span.set_attribute.assert_not_called()


def test_set_llm_output_missing_fields() -> None:
    """It should not fail when missing fields."""
    span = MagicMock()
    _set_llm_output(ChatMessage("assistant"), span)

    span.set_attributes.assert_not_called()


def test_uninstrument_before_instrument() -> None:
    _SmolagentsInstrumentor().uninstrument(MagicMock())



from typing import TYPE_CHECKING, Any

from pydantic import BaseModel

from any_agent.config import AgentConfig, AgentFramework

from .any_agent import AnyAgent

try:
    from agno.agent import Agent
    from agno.models.litellm import LiteLLM
    from agno.tools.toolkit import Toolkit

    DEFAULT_MODEL_TYPE = LiteLLM
    agno_available = True
except ImportError:
    agno_available = False


if TYPE_CHECKING:
    from agno.agent import RunResponse
    from agno.models.base import Model


class AgnoAgent(AnyAgent):
    """Agno agent implementation that handles both loading and running."""

    def __init__(self, config: AgentConfig):
        super().__init__(config)
        self._agent: Agent | None = None

    @property
    def framework(self) -> AgentFramework:
        return AgentFramework.AGNO

    def _get_model(self, agent_config: AgentConfig) -> "Model":
        """Get the model configuration for an Agno agent."""
        model_type = agent_config.model_type or DEFAULT_MODEL_TYPE

        return model_type(
            id=agent_config.model_id,
            api_base=agent_config.api_base,
            api_key=agent_config.api_key,
            request_params=agent_config.model_args or {},  # type: ignore[arg-type]
        )

    @staticmethod
    def _unpack_tools(tools: list[Any]) -> list[Any]:
        unpacked: list[Any] = []
        for tool in tools:
            if isinstance(tool, Toolkit):
                unpacked.extend(f for f in tool.functions.values())
            else:
                unpacked.append(tool)
        return unpacked

    async def _load_agent(self) -> None:
        if not agno_available:
            msg = "You need to `pip install 'any-agent[agno]'` to use this agent"
            raise ImportError(msg)

        tools, _ = await self._load_tools(self.config.tools)

        self._tools = self._unpack_tools(tools)

        agent_args = self.config.agent_args or {}
        if self.config.output_type:
            agent_args["response_model"] = self.config.output_type
        self._agent = Agent(
            name=self.config.name,
            instructions=self.config.instructions,
            model=self._get_model(self.config),
            tools=tools,
            **agent_args,
        )

    async def _run_async(self, prompt: str, **kwargs: Any) -> str | BaseModel:
        if not self._agent:
            error_message = "Agent not loaded. Call load_agent() first."
            raise ValueError(error_message)
        result: RunResponse = await self._agent.arun(prompt, **kwargs)
        return result.content  # type: ignore[return-value]



from typing import TYPE_CHECKING, Any, cast

from pydantic import BaseModel

from any_agent import AgentConfig, AgentFramework
from any_agent.logging import logger
from any_agent.tools.final_output import FinalOutputTool

from .any_agent import AnyAgent

try:
    from llama_index.core.agent.workflow import (
        BaseWorkflowAgent,
        FunctionAgent,
        ReActAgent,
    )
    from llama_index.llms.litellm import LiteLLM

    DEFAULT_AGENT_TYPE = FunctionAgent
    DEFAULT_MODEL_TYPE = LiteLLM
    llama_index_available = True
except ImportError:
    llama_index_available = False


if TYPE_CHECKING:
    from llama_index.core.agent.workflow.workflow_events import AgentOutput
    from llama_index.core.llms import LLM


class LlamaIndexAgent(AnyAgent):
    """LLamaIndex agent implementation that handles both loading and running."""

    def __init__(self, config: AgentConfig):
        super().__init__(config)
        self._agent: BaseWorkflowAgent | None = None

    @property
    def framework(self) -> AgentFramework:
        return AgentFramework.LLAMA_INDEX

    def _get_model(self, agent_config: AgentConfig) -> "LLM":
        """Get the model configuration for a llama_index agent."""
        model_type = agent_config.model_type or DEFAULT_MODEL_TYPE
        return cast(
            "LLM",
            model_type(
                model=agent_config.model_id,
                api_key=agent_config.api_key,
                api_base=agent_config.api_base,
                additional_kwargs=agent_config.model_args or {},  # type: ignore[arg-type]
            ),
        )

    async def _load_agent(self) -> None:
        """Load the LLamaIndex agent with the given configuration."""
        if not llama_index_available:
            msg = "You need to `pip install 'any-agent[llama_index]'` to use this agent"
            raise ImportError(msg)

        instructions = self.config.instructions
        tools_to_use = list(self.config.tools)
        if self.config.output_type:
            instructions = instructions or ""
            instructions += f"""You must return a {self.config.output_type.__name__} JSON string.
            This object must match the following schema:
            {self.config.output_type.model_json_schema()}
            You can use the 'final_output' tool to help verify your output
            """
            tools_to_use.append(FinalOutputTool(self.config.output_type))
        imported_tools, _ = await self._load_tools(tools_to_use)
        agent_type = self.config.agent_type or DEFAULT_AGENT_TYPE
        # if agent type is FunctionAgent but there are no tools, throw an error
        if agent_type == FunctionAgent and not imported_tools:
            logger.warning(
                "FunctionAgent requires tools and none were provided. Using ReActAgent instead."
            )
            agent_type = ReActAgent
        self._tools = imported_tools

        self._agent = agent_type(
            name=self.config.name,
            tools=imported_tools,
            description=self.config.description or "The main agent",
            llm=self._get_model(self.config),
            system_prompt=instructions,
            **self.config.agent_args or {},
        )

    async def _run_async(self, prompt: str, **kwargs: Any) -> str | BaseModel:
        if not self._agent:
            error_message = "Agent not loaded. Call load_agent() first."
            raise ValueError(error_message)
        result: AgentOutput = await self._agent.run(prompt, **kwargs)
        # assert that it's a TextBlock
        if not result.response.blocks or not hasattr(result.response.blocks[0], "text"):
            msg = f"Agent did not return a valid response: {result.response}"
            raise ValueError(msg)
        if self.config.output_type:
            return self.config.output_type.model_validate_json(
                result.response.blocks[0].text
            )
        return result.response.blocks[0].text



from unittest.mock import patch

import pytest
from litellm.utils import validate_environment

from any_agent import (
    AgentConfig,
    AgentFramework,
    AgentRunError,
    AnyAgent,
)
from any_agent.tracing.otel_types import StatusCode


def test_runtime_error(
    agent_framework: AgentFramework,
) -> None:
    """An exception not caught by the framework should be caught by us.

    `AnyAgent.run_async` should catch and reraise an `AgentRunError`.

    The `AgentRunError.trace` should be retrieved.
    """
    kwargs = {}

    kwargs["model_id"] = "gpt-4.1-nano"
    env_check = validate_environment(kwargs["model_id"])
    if not env_check["keys_in_environment"]:
        pytest.skip(f"{env_check['missing_keys']} needed for {agent_framework}")

    model_args = {"temperature": 0.0}

    exc_reason = "It's a trap!"

    patch_function = "litellm.acompletion"
    if agent_framework is AgentFramework.GOOGLE:
        patch_function = "google.adk.models.lite_llm.acompletion"
    elif agent_framework is AgentFramework.SMOLAGENTS:
        patch_function = "litellm.completion"

    with patch(patch_function) as litellm_path:
        litellm_path.side_effect = RuntimeError(exc_reason)
        agent_config = AgentConfig(
            model_id=kwargs["model_id"],
            tools=[],
            model_args=model_args,
        )
        agent = AnyAgent.create(agent_framework, agent_config)
        spans = []
        try:
            agent.run(
                "Write a four-line poem about agent frameworks.",
            )
        except AgentRunError as are:
            spans = are.trace.spans
            assert any(
                span.status.status_code == StatusCode.ERROR
                and span.status.description is not None
                and exc_reason in span.status.description
                for span in spans
            )


def search_web(query: str) -> str:
    """Perform a duckduckgo web search based on your query then returns the top search results.

    Args:
        query (str): The search query to perform.

    Returns:
        The top search results.

    """
    msg = "It's a trap!"
    raise ValueError(msg)


def test_tool_error(
    agent_framework: AgentFramework,
) -> None:
    """An exception raised inside a tool will be caught by us.

    We make sure an appropriate Status is set to the tool execution span.
    We allow the Agent to try to recover from the tool calling failure.
    """
    kwargs = {}

    kwargs["model_id"] = "gpt-4.1-nano"
    env_check = validate_environment(kwargs["model_id"])
    if not env_check["keys_in_environment"]:
        pytest.skip(f"{env_check['missing_keys']} needed for {agent_framework}")

    model_args = {"temperature": 0.0}

    agent_config = AgentConfig(
        model_id=kwargs["model_id"],
        instructions="You must use the available tools to answer questions.",
        tools=[search_web],
        model_args=model_args,
    )

    agent = AnyAgent.create(agent_framework, agent_config)

    agent_trace = agent.run(
        "Check in the web which agent framework is the best.",
    )
    assert any(
        span.is_tool_execution()
        and span.status.status_code == StatusCode.ERROR
        and "It's a trap!" in getattr(span.status, "description", "")
        for span in agent_trace.spans
    )



# mypy: disable-error-code="union-attr"
from __future__ import annotations

import json
from typing import TYPE_CHECKING, Any

from opentelemetry.trace import StatusCode

from .common import _set_tool_output

if TYPE_CHECKING:
    from google.adk.agents.callback_context import CallbackContext
    from google.adk.models.llm_request import LlmRequest
    from google.adk.models.llm_response import LlmResponse
    from google.adk.tools.base_tool import BaseTool
    from google.adk.tools.tool_context import ToolContext
    from opentelemetry.trace import Span

    from any_agent.frameworks.google import GoogleAgent


def _set_llm_output(llm_response: LlmResponse, span: Span) -> None:
    content = llm_response.content
    if not content:
        return
    if not content.parts:
        return

    if content.parts[0].text:
        span.set_attributes(
            {
                "gen_ai.output": content.parts[0].text,
                "gen_ai.output.type": "text",
            }
        )
    else:
        span.set_attributes(
            {
                "gen_ai.output": json.dumps(
                    [
                        {
                            "tool.name": getattr(part.function_call, "name", "No name"),
                            "tool.args": getattr(part.function_call, "args", "No args"),
                        }
                        for part in content.parts
                        if part.function_call
                    ],
                    default=str,
                    ensure_ascii=False,
                ),
                "gen_ai.output.type": "json",
            }
        )


def _set_llm_input(llm_request: LlmRequest, span: Span) -> None:
    if not llm_request.contents:
        return
    messages = []
    if config := llm_request.config:
        messages.append(
            {
                "role": "system",
                "content": getattr(config, "system_instruction", "No instructions"),
            }
        )
    if parts := llm_request.contents[0].parts:
        messages.append(
            {
                "role": getattr(llm_request.contents[0], "role", "No role"),
                "content": getattr(parts[0], "text", "No content"),
            }
        )
    span.set_attribute(
        "gen_ai.input.messages",
        json.dumps(messages, default=str, ensure_ascii=False),
    )


class _GoogleADKInstrumentor:
    def __init__(self) -> None:
        self._original: dict[str, Any] = {}
        self.first_llm_calls: set[int] = set()
        self._current_spans: dict[str, dict[str, Span]] = {
            "model": {},
            "tool": {},
        }

    def instrument(self, agent: GoogleAgent) -> None:
        if len(agent._running_traces) > 1:
            return

        tracer = agent._tracer

        self._original["before_model"] = agent._agent.before_model_callback

        def before_model_callback(
            callback_context: CallbackContext,
            llm_request: LlmRequest,
        ) -> Any | None:
            span: Span = tracer.start_span(
                name=f"call_llm {llm_request.model}",
            )
            span.set_attributes(
                {
                    "gen_ai.operation.name": "call_llm",
                    "gen_ai.request.model": llm_request.model or "no_model",
                }
            )
            trace_id = span.get_span_context().trace_id
            if trace_id not in self.first_llm_calls:
                self.first_llm_calls.add(trace_id)
                _set_llm_input(llm_request, span)
            self._current_spans["model"][callback_context.invocation_id] = span

            if callable(self._original["before_model"]):
                return self._original["before_model"](callback_context, llm_request)

            return None

        agent._agent.before_model_callback = before_model_callback

        self._original["before_tool"] = agent._agent.before_tool_callback

        def before_tool_callback(
            tool: BaseTool,
            args: dict[str, Any],
            tool_context: ToolContext,
        ) -> Any | None:
            span: Span = tracer.start_span(
                name=f"execute_tool {tool.name}",
            )
            span.set_attributes(
                {
                    "gen_ai.operation.name": "execute_tool",
                    "gen_ai.tool.name": tool.name,
                    "gen_ai.tool.description": tool.description,
                    "gen_ai.tool.args": json.dumps(args),
                    "gen_ai.tool.call.id": getattr(
                        tool_context, "function_call_id", "no_id"
                    ),
                }
            )

            self._current_spans["tool"][tool_context.invocation_id] = span

            if callable(self._original["before_tool"]):
                return self._original["before_tool"](tool, args, tool_context)

            return None

        agent._agent.before_tool_callback = before_tool_callback

        self._original["after_model"] = agent._agent.after_model_callback

        def after_model_callback(
            callback_context: CallbackContext,
            llm_response: LlmResponse,
        ) -> Any | None:
            span = self._current_spans["model"][callback_context.invocation_id]

            _set_llm_output(llm_response, span)
            if resp_meta := llm_response.usage_metadata:
                if prompt_tokens := resp_meta.prompt_token_count:
                    span.set_attributes({"gen_ai.usage.input_tokens": prompt_tokens})
                if output_tokens := resp_meta.candidates_token_count:
                    span.set_attributes({"gen_ai.usage.output_tokens": output_tokens})
            span.set_status(StatusCode.OK)
            span.end()
            trace_id = span.get_span_context().trace_id
            agent._running_traces[trace_id].add_span(span)

            del self._current_spans["model"][callback_context.invocation_id]

            if callable(self._original["after_model"]):
                return self._original["after_model"](callback_context, llm_response)

            return None

        agent._agent.after_model_callback = after_model_callback

        self._original["after_tool"] = agent._agent.after_tool_callback

        def after_tool_callback(
            tool: BaseTool,
            args: dict[str, Any],
            tool_context: ToolContext,
            tool_response: dict[Any, Any],
        ) -> Any | None:
            span = self._current_spans["tool"][tool_context.invocation_id]

            _set_tool_output(tool_response, span)
            span.end()

            trace_id = span.get_span_context().trace_id
            agent._running_traces[trace_id].add_span(span)

            del self._current_spans["tool"][tool_context.invocation_id]

            if callable(self._original["after_tool"]):
                return self._original["after_tool"](
                    tool, args, tool_context, tool_response
                )

            return None

        agent._agent.after_tool_callback = after_tool_callback

    def uninstrument(self, agent: GoogleAgent) -> None:
        if len(agent._running_traces) > 1:
            return
        if "before_model" in self._original:
            agent._agent.before_model_callback = self._original["before_model"]
        if "before_tool" in self._original:
            agent._agent.before_tool_callback = self._original["before_tool"]
        if "after_model" in self._original:
            agent._agent.after_model_callback = self._original["after_model"]
        if "after_tool" in self._original:
            agent._agent.after_tool_callback = self._original["after_tool"]



from collections.abc import Callable
from typing import Any

from pydantic import BaseModel

from any_agent.tracing.agent_trace import AgentTrace


class TraceTools:
    def __init__(self, trace: AgentTrace):
        self.trace = trace

    def get_all_tools(self) -> list[Callable[[], Any]]:
        """Get all tool functions from this class.

        Returns:
            list[callable]: List of all tool functions

        """
        # Get all methods that don't start with underscore and aren't get_all_tools
        tools = []
        for attr_name in dir(self):
            if not attr_name.startswith("_") and attr_name != "get_all_tools":
                attr = getattr(self, attr_name)
                if callable(attr) and attr_name not in ["trace"]:
                    tools.append(attr)
        return tools

    def get_final_output(self) -> str | BaseModel | None:
        """Get the final output from the agent trace.

        Returns:
            str | BaseModel | None: The final output of the agent

        """
        return self.trace.final_output

    def get_tokens_used(self) -> int:
        """Get the number of tokens used by the agent as reported by the trace.

        Returns:
            int: The number of tokens used by the agent

        """
        return self.trace.tokens.total_tokens

    def get_steps_taken(self) -> int:
        """Get the number of steps taken by the agent as reported by the trace.

        Returns:
            int: The number of steps taken by the agent

        """
        return len(self.trace.spans)

    def get_messages_from_trace(self) -> str:
        """Get a summary of what happened in each step/span of the agent trace.

        This includes information about the input, output, and tool calls for each step.

        Returns:
            str: The evidence of all the spans in the trace

        """
        messages = self.trace.spans_to_messages()
        evidence = ""
        for message in messages:
            evidence += f"### {message.role}\n{message.content}\n\n"
        return evidence

    def get_duration(self) -> float:
        """Get the duration of the agent trace.

        Returns:
            float: The duration in seconds of the agent trace

        """
        return self.trace.duration.total_seconds()



from collections.abc import Callable, Mapping, MutableMapping, Sequence
from enum import StrEnum, auto
from typing import Any, Self

from pydantic import BaseModel, ConfigDict, Field


class AgentFramework(StrEnum):
    GOOGLE = auto()
    LANGCHAIN = auto()
    LLAMA_INDEX = auto()
    OPENAI = auto()
    AGNO = auto()
    SMOLAGENTS = auto()
    TINYAGENT = auto()

    @classmethod
    def from_string(cls, value: str | Self) -> Self:
        if isinstance(value, cls):
            return value

        formatted_value = value.strip().upper()
        if formatted_value not in cls.__members__:
            error_message = (
                f"Unsupported agent framework: '{value}'. "
                f"Valid frameworks are: {list(cls.__members__.keys())}"
            )
            raise ValueError(error_message)

        return cls[formatted_value]


class MCPStdio(BaseModel):
    command: str
    """The executable to run to start the server.

    For example, `docker`, `uvx`, `npx`.
    """

    args: Sequence[str]
    """Command line args to pass to the command executable.

    For example, `["run", "-i", "--rm", "mcp/fetch"]`.
    """

    env: dict[str, str] | None = None
    """The environment variables to set for the server."""

    tools: Sequence[str] | None = None
    """List of tool names to use from the MCP Server.

    Use it to limit the tools accessible by the agent.
    For example, if you use [`mcp/filesystem`](https://hub.docker.com/r/mcp/filesystem),
    you can pass `tools=["read_file", "list_directory"]` to limit the agent to read-only operations.

    If none is specified, the default behavior is that the agent will have access to all tools under that MCP server.
    """

    client_session_timeout_seconds: float | None = 5
    """the read timeout passed to the MCP ClientSession."""

    model_config = ConfigDict(frozen=True, extra="forbid")


class MCPSse(BaseModel):
    url: str
    """The URL of the server."""

    headers: Mapping[str, str] | None = None
    """The headers to send to the server."""

    tools: Sequence[str] | None = None
    """List of tool names to use from the MCP Server.

    Use it to limit the tools accessible by the agent.
    For example, if you use [`mcp/filesystem`](https://hub.docker.com/r/mcp/filesystem),
    you can pass `tools=["read_file", "list_directory"]` to limit the agent to read-only operations.
    """

    client_session_timeout_seconds: float | None = 5
    """the read timeout passed to the MCP ClientSession."""

    model_config = ConfigDict(frozen=True)


class ServingConfig(BaseModel):
    """Configuration for serving an agent using the Agent2Agent Protocol (A2A).

    We use the example `A2ASever` from https://github.com/google/A2A/tree/main/samples/python.
    """

    model_config = ConfigDict(extra="forbid")

    host: str = "localhost"
    """Will be passed as argument to `uvicorn.run`."""

    port: int = 5000
    """Will be passed as argument to `uvicorn.run`."""

    endpoint: str = "/"
    """Will be pass as argument to `Starlette().add_route`"""

    log_level: str = "warning"
    """Will be passed as argument to the `uvicorn` server."""

    version: str = "0.1.0"


MCPParams = MCPStdio | MCPSse

Tool = str | MCPParams | Callable[..., Any]


class AgentConfig(BaseModel):
    model_config = ConfigDict(extra="forbid")

    model_id: str
    """Select the underlying model used by the agent.

    If you are using the default model_type (LiteLLM), you can refer to [LiteLLM Provider Docs](https://docs.litellm.ai/docs/providers) for the list of providers and how to access them.
    """

    api_base: str | None = None
    api_key: str | None = None

    description: str | None = None
    """Description of the agent."""

    name: str = "any_agent"
    """The name of the agent.

    Defaults to `any_agent`.
    """

    instructions: str | None = None
    """Specify the instructions for the agent (often also referred to as a `system_prompt`)."""

    tools: Sequence[Tool] = Field(default_factory=list)
    """List of tools to be used by the agent.

    See more info at [Tools](../agents/tools.md).
    """

    agent_type: Callable[..., Any] | None = None
    """Control the type of agent class that is used by the framework, and is unique to the framework used.

    Check the individual `Frameworks` pages for more info on the defaults.
    """

    agent_args: MutableMapping[str, Any] | None = None
    """Pass arguments to the instance used by the underlying framework.

    For example, you can pass `output_type` when using the OpenAI Agents SDK:

    ```py
    from pydantic import BaseModel

    class CalendarEvent(BaseModel):
        name: str
        date: str
        participants: list[str]

    agent = AnyAgent.create(
        AgentConfig(
            model_id="gpt-4.1-mini",
            instructions="Extract calendar events from text",
            agent_args={
                "output_type": CalendarEvent
            }
        )
    )
    ```
    """

    model_type: Callable[..., Any] | None = None
    """Control the type of model class that is used by the agent framework, and is unique to the agent framework being used.

    For each framework, we leverage their support for LiteLLM and use it as default model_type, allowing you to use the same model_id syntax across these frameworks.
    """

    model_args: MutableMapping[str, Any] | None = None
    """Pass arguments to the model instance like `temperature`, `top_k`, as well as any other provider-specific parameters.

    Refer to LiteLLM Completion API Docs for more info.
    """

    output_type: type[BaseModel] | None = None
    """Control the output schema from calling `run`. By default, the agent will return a type str.

    Using this parameter you can define a Pydantic model that will be returned by the agent run methods.
    """



# mypy: disable-error-code="method-assign, union-attr"
from __future__ import annotations

import json
from typing import TYPE_CHECKING, Any

from opentelemetry.trace import StatusCode

from .common import _set_tool_output

if TYPE_CHECKING:
    from uuid import UUID

    from langchain_core.messages import BaseMessage
    from langchain_core.outputs import LLMResult
    from opentelemetry.trace import Span, Tracer

    from any_agent import AgentTrace
    from any_agent.frameworks.langchain import LangchainAgent


def _set_llm_input(messages: list[list[BaseMessage]], span: Span) -> None:
    if not messages:
        return
    if not messages[0]:
        return

    span.set_attribute(
        "gen_ai.input.messages",
        json.dumps(
            [
                {
                    "role": message.type.replace("human", "user"),
                    "content": message.content,
                }
                for message in messages[0]
            ],
            default=str,
            ensure_ascii=False,
        ),
    )


def _set_llm_output(response: LLMResult, span: Span) -> None:
    if not response.generations:
        return
    if not response.generations[0]:
        return

    generation = response.generations[0][0]

    if text := generation.text:
        span.set_attributes(
            {
                "gen_ai.output": text,
                "gen_ai.output.type": "text",
            }
        )
    if message := getattr(generation, "message", None):
        if tool_calls := getattr(message, "tool_calls", None):
            span.set_attributes(
                {
                    "gen_ai.output": json.dumps(
                        [
                            {
                                "tool.name": tool.get("name", "No name"),
                                "tool.args": tool.get("args", "No args"),
                            }
                            for tool in tool_calls
                        ],
                        default=str,
                        ensure_ascii=False,
                    ),
                    "gen_ai.output.type": "json",
                }
            )

    if llm_output := getattr(response, "llm_output", None):
        if token_usage := llm_output.get("token_usage", None):
            span.set_attributes(
                {
                    "gen_ai.usage.input_tokens": token_usage.prompt_tokens,
                    "gen_ai.usage.output_tokens": token_usage.completion_tokens,
                }
            )


class _LangChainInstrumentor:
    def __init__(self) -> None:
        self._original_ainvoke: Any | None = None

    def instrument(self, agent: LangchainAgent) -> None:
        if len(agent._running_traces) > 1:
            return

        from langchain_core.callbacks.base import BaseCallbackHandler
        from langchain_core.runnables import RunnableConfig

        class _LangChainTracingCallback(BaseCallbackHandler):
            def __init__(
                self, tracer: Tracer, running_traces: dict[int, AgentTrace]
            ) -> None:
                self.tracer = tracer
                self.running_traces = running_traces
                self._current_spans: dict[str, dict[str, Span]] = {
                    "model": {},
                    "tool": {},
                }
                self.first_llm_calls: set[int] = set()
                super().__init__()

            def on_chat_model_start(
                self,
                serialized: dict[str, Any],
                messages: list[list[BaseMessage]],
                *,
                run_id: UUID,
                parent_run_id: UUID | None = None,
                tags: list[str] | None = None,
                metadata: dict[str, Any] | None = None,
                **kwargs: Any,
            ) -> Any:
                model = kwargs.get("invocation_params", {}).get("model", "No model")
                span: Span = self.tracer.start_span(
                    name=f"call_llm {model}",
                )
                span.set_attributes(
                    {
                        "gen_ai.operation.name": "call_llm",
                        "gen_ai.request.model": model,
                    }
                )
                trace_id = span.get_span_context().trace_id
                if trace_id not in self.first_llm_calls:
                    self.first_llm_calls.add(trace_id)
                    _set_llm_input(messages, span)

                self._current_spans["model"][str(run_id)] = span

            def on_tool_start(
                self,
                serialized: dict[str, Any],
                input_str: str,
                *,
                run_id: UUID,
                parent_run_id: UUID | None = None,
                tags: list[str] | None = None,
                metadata: dict[str, Any] | None = None,
                inputs: dict[str, Any] | None = None,
                **kwargs: Any,
            ) -> Any:
                span: Span = self.tracer.start_span(
                    name=f"execute_tool {serialized.get('name')}",
                )
                span.set_attributes(
                    {
                        "gen_ai.operation.name": "execute_tool",
                        "gen_ai.tool.name": serialized.get("name", "No name"),
                        "gen_ai.tool.description": serialized.get(
                            "description", "No description"
                        ),
                        "gen_ai.tool.args": json.dumps(
                            inputs, default=str, ensure_ascii=False
                        ),
                    }
                )

                self._current_spans["tool"][str(run_id)] = span

            def on_llm_end(
                self,
                response: LLMResult,
                *,
                run_id: UUID,
                parent_run_id: UUID | None = None,
                **kwargs: Any,
            ) -> Any:
                span = self._current_spans["model"][str(run_id)]
                _set_llm_output(response, span)
                span.set_status(StatusCode.OK)
                span.end()

                trace_id = span.get_span_context().trace_id
                self.running_traces[trace_id].add_span(span)
                del self._current_spans["model"][str(run_id)]

            def on_tool_end(
                self,
                output: Any,
                *,
                run_id: UUID,
                parent_run_id: UUID | None = None,
                **kwargs: Any,
            ) -> Any:
                span = self._current_spans["tool"][str(run_id)]

                if content := getattr(output, "content", None):
                    _set_tool_output(content, span)
                    if tool_call_id := getattr(output, "tool_call_id", None):
                        span.set_attribute("gen_ai.tool.call.id", tool_call_id)

                span.end()
                trace_id = span.get_span_context().trace_id
                self.running_traces[trace_id].add_span(span)
                del self._current_spans["tool"][str(run_id)]

        tracing_callback = _LangChainTracingCallback(
            agent._tracer, agent._running_traces
        )

        self._original_ainvoke = agent._agent.ainvoke

        async def wrap_ainvoke(*args, **kwargs):  # type: ignore[no-untyped-def]
            if "config" in kwargs:
                if callbacks := kwargs["config"].get("callbacks"):
                    if isinstance(callbacks, list):
                        kwargs["config"]["callbacks"].append(tracing_callback)
                    else:
                        original_callback = kwargs["config"]["callbacks"]
                        kwargs["config"]["callbacks"] = [
                            original_callback,
                            tracing_callback,
                        ]
                else:
                    kwargs["config"]["callbacks"] = [tracing_callback]
            else:
                kwargs["config"] = RunnableConfig(callbacks=[tracing_callback])

            return await self._original_ainvoke(*args, **kwargs)  # type: ignore[misc]

        agent._agent.ainvoke = wrap_ainvoke

    def uninstrument(self, agent: LangchainAgent) -> None:
        if len(agent._running_traces) > 1:
            return
        if self._original_ainvoke is not None:
            agent._agent.ainvoke = self._original_ainvoke



# mypy: disable-error-code="method-assign,no-untyped-def,union-attr"
from __future__ import annotations

import json
from typing import TYPE_CHECKING, Any

from opentelemetry.trace import StatusCode

from .common import _set_tool_output

if TYPE_CHECKING:
    from agno.models.message import Message, MessageMetrics
    from agno.models.response import ModelResponse
    from opentelemetry.trace import Span

    from any_agent.frameworks.agno import AgnoAgent


def _set_llm_input(messages: list[Message], span: Span) -> None:
    if not messages:
        return
    span.set_attribute(
        "gen_ai.input.messages",
        json.dumps(
            [
                {"role": message.role, "content": message.content}
                for message in messages
            ],
            default=str,
            ensure_ascii=False,
        ),
    )


def _set_llm_output(assistant_message: Message, span: Span) -> None:
    if content := getattr(assistant_message, "content", None):
        span.set_attributes(
            {
                "gen_ai.output": str(content),
                "gen_ai.output.type": "text",
            }
        )
    if tool_calls := getattr(assistant_message, "tool_calls", None):
        span.set_attributes(
            {
                "gen_ai.output": json.dumps(
                    [
                        {
                            "tool.name": tool.get("function", {}).get(
                                "name", "No name"
                            ),
                            "tool.args": tool.get("function", {}).get(
                                "arguments", "No args"
                            ),
                        }
                        for tool in tool_calls
                    ],
                    default=str,
                    ensure_ascii=False,
                ),
                "gen_ai.output.type": "json",
            }
        )
    metrics: MessageMetrics | None
    if metrics := getattr(assistant_message, "metrics", None):
        span.set_attributes(
            {
                "gen_ai.usage.input_tokens": metrics.input_tokens,
                "gen_ai.usage.output_tokens": metrics.output_tokens,
            }
        )


class _AgnoInstrumentor:
    def __init__(self) -> None:
        self._original_aprocess_model: Any = None
        self._original_arun_function_calls: Any = None
        self.first_llm_calls: set[int] = set()

    def instrument(self, agent: AgnoAgent) -> None:
        if len(agent._running_traces) > 1:
            return

        model = agent._agent.model
        tracer = agent._tracer

        self._original_aprocess_model = model._aprocess_model_response

        async def wrap_aprocess_model_response(
            *args,
            **kwargs,
        ) -> tuple[Message, bool]:
            with tracer.start_as_current_span(f"call_llm {model.id}") as span:
                span.set_attributes(
                    {
                        "gen_ai.operation.name": "call_llm",
                        "gen_ai.request.model": model.id,
                    }
                )
                trace_id = span.get_span_context().trace_id
                if trace_id not in self.first_llm_calls:
                    self.first_llm_calls.add(trace_id)
                    _set_llm_input(kwargs.get("messages", []), span)

                assistant_message: Message
                has_tool_calls: bool
                assistant_message, has_tool_calls = await self._original_aprocess_model(
                    *args, **kwargs
                )

                _set_llm_output(assistant_message, span)

                span.set_status(StatusCode.OK)

            agent._running_traces[trace_id].add_span(span)
            return assistant_message, has_tool_calls

        model._aprocess_model_response = wrap_aprocess_model_response

        self._original_arun_function_calls = model.arun_function_calls

        async def wrap_arun_function_calls(*args, **kwargs):
            tool_call_spans = {}
            function_call_response: ModelResponse
            async for function_call_response in self._original_arun_function_calls(
                *args, **kwargs
            ):
                if function_call_response.event == "ToolCallStarted":
                    if tool_executions := function_call_response.tool_executions:
                        tool = function_call_response.tool_executions[0]
                        tool_name = getattr(tool, "tool_name", "No name")
                        tool_args = getattr(tool, "tool_args", {})
                        tool_call_id = getattr(tool, "tool_call_id", "No id")
                        span: Span = tracer.start_span(
                            name=f"execute_tool {tool_name}",
                        )
                        span.set_attributes(
                            {
                                "gen_ai.operation.name": "execute_tool",
                                "gen_ai.tool.name": tool_name,
                                "gen_ai.tool.args": json.dumps(
                                    tool_args,
                                    default=str,
                                    ensure_ascii=False,
                                ),
                                "gen_ai.tool.call.id": tool_call_id,
                            }
                        )
                    tool_call_spans[tool_call_id] = span
                elif function_call_response.event == "ToolCallCompleted":
                    if tool_executions := function_call_response.tool_executions:
                        tool_call_id = getattr(
                            tool_executions[0], "tool_call_id", "No id"
                        )
                        span = tool_call_spans[tool_call_id]
                        _set_tool_output(
                            getattr(tool_executions[0], "result", "{}"), span
                        )

                    span.end()
                    trace_id = span.get_span_context().trace_id
                    agent._running_traces[trace_id].add_span(span)
                yield function_call_response

        model.arun_function_calls = wrap_arun_function_calls

    def uninstrument(self, agent: AgnoAgent):
        if len(agent._running_traces) > 1:
            return
        model = agent._agent.model
        if self._original_aprocess_model is not None:
            model._aprocess_model_response = self._original_aprocess_model
        if self._original_arun_function_calls is not None:
            model.arun_function_calls = self._original_arun_function_calls
